@@ -25,6 +25,7 @@
 
 #include <stdbool.h>
 #include <stdint.h>
+#include <string.h>
 
 struct in_addr;
 struct in6_addr;
@@ -40,6 +41,26 @@ bool l_net_hostname_is_localhost(const char *hostname);
 bool l_net_get_address(int ifindex, struct in_addr *out);
 bool l_net_get_link_local_address(int ifindex, struct in6_addr *out);
 
+static inline bool l_net_subnet_matches(const void *a, const void *b,
+					uint8_t prefix_len)
+{
+	uint8_t bytes = prefix_len / 8;
+	uint8_t bits = prefix_len & 7;
+	uint8_t left = ((const uint8_t *) a)[bytes];
+	uint8_t right = ((const uint8_t *) b)[bytes];
+
+	/*
+	 * @a and @b are network byte order IPv4 or IPv6 addresses.
+	 * We want to check if the initial (top) @prefix_len bits match.
+	 * memcmp the whole bytes, then compare the final byte's top
+	 * bits by anding with a mask.
+	 */
+	if (memcmp(a, b, bytes))
+		return false;
+
+	return !bits || ((left ^ right) & (0xff00u >> bits)) == 0;
+}
+
 #ifdef __cplusplus
 }
 #endif
